"""
Preprocessing routines (:mod:`stnmf.preprocessing`)
===================================================

Data is brought into appropriate format to be used with and passed to
:class:`stnmf.STNMF`.

.. automodule:: stnmf.preprocessing.spiketriggered
  :no-members:
  :no-inherited-members:
  :no-special-members:

"""
__all__ = [
    'spiketriggered'
]

from . import spiketriggered
